/*****************************************************************************
*   $Id: dgdb.c,v 1.1.1.1 2000/06/22 15:38:22 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgdb.c,v $
* Revision 1.1.1.1  2000/06/22 15:38:22  tekno
* Opened source
*
*
*****************************************************************************/

/* dgdb.c - read and understand a COOK database file.  
   Also API for programmatic creation of database or associative array. */


#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <setjmp.h>
#include <stddef.h>

/*#define DEBUG*/

#include "dgdb.h"

extern int verbose;

static FILE * db;
static jmp_buf jb;
static long lineno;
static char * current_file;

/* Externally visible vars for other modules */
hash_table * recht;
int nrec;
record ** rpt;

/* Single-char record recognition */
static record * recognition_1_tab[128];
/* Double-char record recognition, indexed by 1st char */
static record ** recognition_2_tab[128];

#define DBERROR(s) (longjmp(jb, (int)(s)))
#define DBWARNING(s) dbwarning(s);
#define NEW(type) ((type *)dbcalloc(sizeof(type)))

static void dbwarning(char * s)
{
  fprintf(stderr, "%s:%d: %s\n",
      current_file,
      lineno,
      s);
}

static char * dbcalloc(int size)
{
  char * r = (char *)calloc(size, 1);
  if (!r)
    DBERROR("Virtual memory shortage");
  return r;
}

static char * dbmalloc(int size)
{
  char * r = (char *)malloc(size);
  if (!r)
    DBERROR("Virtual memory shortage");
  return r;
}

static char * char_stack;
static char * stack_top;
static long chunksize;
static char * char_start;
static char * last_char_start;
#define STACK_CHUNK 65536

static void push_char(int c)
{
  if (stack_top - char_stack == chunksize)
  {
    if (char_start == char_stack)
    {
      char_stack = realloc(char_stack, chunksize + STACK_CHUNK);
      if (!char_stack)
        DBERROR("Virtual memory shortage");
      stack_top = char_stack + chunksize;
      char_start = char_stack;
    }
    else
    {
      char_stack = dbmalloc(chunksize + STACK_CHUNK);
      memcpy(char_stack, char_start, stack_top - char_start);
      stack_top = stack_top - char_start + char_stack;
      char_start = char_stack;
    }
    chunksize += STACK_CHUNK;
  }
  *stack_top++ = c;
}

static char * finish_push(void)
{
  char * r = char_start;
  push_char(0);
  char_start = stack_top;
  last_char_start = r;
  return r;
}

static void discard_push(void)
{
  stack_top = char_start;
}

static void discard_last(void)
{
  if (last_char_start)
  {
    stack_top = char_start = last_char_start;
    last_char_start = 0;
  }
}

static char * makestr(int len)
{
  /* Allocate string (len+1 chars) */
  char * r;
  
  r = (char *)malloc(len+1);
  if (!r)
    DBERROR("Virtual memory shortage");
  return r;
}

static int sline;	/* Start line of current field */

static char * read_to_eol(int * special, char nocopy)
{
  char * r = 0;
  int c;
  int crcount = 0;
  
  if (special) *special = 0;
  for (;;)
  {
    c = getc(db);
    if (c == EOF)
    {
      fprintf(stderr, "from starting line %d...\n", sline);
      DBERROR("Missing string field");
    }
    if (c == '\n')
    {
      if (crcount)
      {
        ungetc(c, db);	/* Empty string */
        return finish_push();
      }
      crcount++;
      lineno++;
    }
    if (!isspace(c))
      break;
  }
  if (!nocopy && (c == '*' || c == '$'))
  {
    push_char(c);
    if (special) *special = 1;
    while ((c = getc(db)) != '\n' && c != EOF);
  }
  else for (;;)
  {
    push_char(c);
    c = getc(db);
    if (c == '\n' || c == EOF)
      break;
  }
  if (c == '\n')
    ungetc(c, db);
  return finish_push();
}

static char * read_stringfield(int * special, char nocopy)
{
  char * r = 0;
  int c;
  char clo, opn;
  int nest;
  
  sline = lineno;
  
  if (special) *special = 0;
  
  for (;;)
  {
    c = getc(db);
    if (c == EOF)
    {
      fprintf(stderr, "from starting line %d...\n", sline);
      DBERROR("Missing string field");
    }
    if (c == '\n')
      lineno++;
    if (!isspace(c))
      break;
  }
  
  switch (c)
  {
  case '*':
  case '$':
    if (nocopy)
      goto normal;
    push_char(c);
    if (special) *special = 1;
    break;
  case '[' : clo = ']'; goto read_bracketed;
  case '(' : clo = ')'; goto read_bracketed;
  case '{' : clo = '}'; goto read_bracketed;
  case '<' : clo = '>'; goto read_bracketed;
  case '`' : clo = '\'';
  read_bracketed:
    for (;;)
    {
      c = getc(db);
      if (c == EOF)
      {
        fprintf(stderr, "from starting line %d...\n", sline);
        DBERROR("End of file while reading string field");
      }
      if (c == clo)
        break;
      else if (c == '\n')
        lineno++;
      else if (c == '\\')
      {
        c = getc(db);
        if (c != '\\' && c != clo)
        {
          ungetc(c, db);
          c = '\\';
        }
      }
      push_char(c);
    }
    break;
  default:
  normal:
    nest = 0;
    for(;;)
    {
      push_char(c);
      c = getc(db);
      if (c == EOF)
      {
        fprintf(stderr, "from starting line %d...\n", sline);
        DBERROR("End of file while reading string field");
      }
      if (!nest)
      {
        if (isspace(c))
        {
          if (c == '\n')
            ungetc(c, db);
          break;
        }
        switch (c)
        {
        case '[' : clo = ']'; goto set_opn;
        case '(' : clo = ')'; goto set_opn;
        case '{' : clo = '}'; goto set_opn;
        case '<' : clo = '>'; goto set_opn;
        case '`' : clo = '\'';
  set_opn:
          opn = c;
          nest++;
        }
      }
      else
      {
        if (c == clo)
          nest--;
        else if (c == opn)
          nest++;
        else if (c == '\n')
          lineno++;
      }
    }
    break;
  }

bye:
  return finish_push();
}


static unsigned compute_hash(char * s, unsigned modulo, int len)
{
  unsigned long hh = 0;
  int ll = len;
  
  if (!s)
    return 0;
  while (len ? ll-- : *s)
  {
    if (hh & 0x80000000) 
      hh = hh << 1 ^ 3;
    else
      hh <<= len ? 8 : 1;
    hh ^= (unsigned char)*s++;
  }
  return (unsigned)(hh % modulo);
}


#define HASH_INDEX(ht, data) \
  compute_hash((ht)->instr ? (data) + (ht)->offs : \
                             *(char **)((data) + (ht)->offs), \
               (ht)->slots, (ht)->instr ? sizeof(int) : 0)

#define MATCH_KEY(ht, s, data) \
  !strcmp((ht)->instr ? (data) + (ht)->offs : \
                        *(char **)((data) + (ht)->offs), \
          (s))

void add_hash(hash_table * ht, char * data)
{
  unsigned hh = HASH_INDEX(ht, data);
  ht_entry * f, * e = ht->htes + hh;
  if (e->data)
  {
    f = NEW(ht_entry);
    memcpy(f, e, sizeof(*f));
    e->next = f;
  }
  e->data = data;
  ht->nent++;
}

void unhash(hash_table * ht, char * data)
{
  unsigned hh = HASH_INDEX(ht, data);
  ht_entry * f, * e = ht->htes + hh;
  if (!e->data)
    return;
  if (e->data == data)
  {
    if ((f = e->next))
    {
      memcpy(e, f, sizeof(ht_entry));
      free(f);
    }
    else
      e->data = 0;
    ht->nent--;
    return;
  }
  for (; e->next; e = e->next)
    if (e->next->data == data)
    {
      f = e->next;
      e->next = f->next;
      free(f);
      ht->nent--;
    }
}

void delete_ht(hash_table ** htp, void (*callback)(char *))
{
  /* Delete the hash table pointed to by *htp.  Zero the
     pointer.  For each data item found, call the given callback. */
  hash_table * ht;
  ht_entry * e, * f, * g;
  if (!*htp)
    return;
  ht = *htp;
  for (e = ht->htes; ht->slots; e++, ht->slots--)
  {
    if (callback && e->data)
      callback(e->data);
    for (f = e->next; f; f = g)
    {
      if (callback && f->data)
        callback(f->data);
      g = f->next;
      free(f);
    }
  }
  free(ht->htes);
  free(ht);
  *htp = 0;
}

hash_table * init_ht(hash_table * ht, unsigned slots, int offs)
{
  if (!ht)
    ht = (hash_table *)malloc(sizeof(hash_table));
  ht->slots = slots;
  ht->offs = offs;
  ht->nent = 0;
  ht->instr = 0;
  ht->htes = (ht_entry *)calloc(sizeof(ht_entry) * ht->slots, 1);
  return ht;
}

char * lookup_hash(hash_table * ht, char * s)
{
  /* Find entry matching string 's'.  If found, return data pointer,
     else null. */
  unsigned hh = compute_hash(s, ht->slots, ht->instr ? sizeof(int) : 0);
  ht_entry * e = ht->htes + hh;
  if (!e->data)
    return 0;
  if (MATCH_KEY(ht, s, e->data))
    return e->data;
  for (e = e->next; e; e = e->next)
    if (MATCH_KEY(ht, s, e->data))
      return e->data;
  return 0;
}

void rehash_table(hash_table * ht, unsigned sl)
{
  /* Hash table initialisation or rehashing to different number of slots */
  if (!ht->nent)
  {
    if (ht->htes)
      free(ht->htes);
    ht->htes = (ht_entry *)dbcalloc(sizeof(ht_entry) * sl);
    ht->slots = sl;
  }
  else
  {
    /* Currently got entries, therefore rehashing */
    unsigned i, oldslots;
    ht_entry * e;
    ht_entry * hte = ht->htes;
    ht->htes = (ht_entry *)dbcalloc(sizeof(ht_entry) * sl);
    oldslots = ht->slots;
    ht->slots = sl;
    for (i = 0, e = hte; i < oldslots; i++, e++)
      while (e->data)
      {
        ADD_HASH(ht, e->data);
        e = e->next;
      }
    free(hte);
  }
}


static field * all_fields;
static record * all_records;

#define RBN_SLOTS 127
static hash_table record_by_name;	/* All records */
#define FBN_SLOTS 477
static hash_table field_by_name;	/* Master fields */
#define OBJ_SLOTS 477			/* Default number of slots for objects*/

record * find_record(char * name)
{
  return (record *)lookup_hash(&record_by_name, name);
}

field * find_master_field(char * name)
{
  return (field *)lookup_hash(&field_by_name, name);
}

static int valid_name(char * n)
{
  if (!isalpha(*n))
    return 0;
  for(n++; *n; n++)
    if (!isalnum(*n) && *n != '_')
      return 0;
  return 1;
}

record * create_record_type(hash_table * recbyname, char * name,
                            int nfields, ... /* field names and types */)
{
  record * r;
  va_list ap;
  int fn;
  field * f = 0;
  unsigned flgs;
  
  va_start(ap, nfields);
  if (setjmp(jb))
  {
    return 0;
  }
  r = NEW(record);
  r->name = name;
  r->nfields = nfields;
  r->definitive = -1;
  /* varargs are in pairs [unsigned, char *] where unsigned contains flags
     selected from CRT_** defines, and char * is field name */
  r->fields = (field *)dbcalloc(sizeof(field) * nfields);
  for (fn = 0, f = r->fields; fn < nfields; fn++, f++)
  {
    flgs = va_arg(ap, unsigned);
    f->name = va_arg(ap, char *);
    f->refname = f->name;
    f->owner = r;
    f->fldno = fn;
    f->type = flgs & CRT_INT ? FT_INT : FT_STRING;
    f->master = f;
    f->definitive = !!(flgs & CRT_DEFINITIVE);
    if (f->definitive)
      r->definitive = fn;
  }
  r->objsize = sizeof(object) + sizeof(char *) * 
                 (r->nfields - 1 + (r->definitive >= 0));
  r->objht.slots = OBJ_SLOTS;
  r->objht.offs = sizeof(object) + sizeof(char *) * (r->definitive - 1);
  r->objht.htes = (ht_entry *)dbcalloc(sizeof(ht_entry) * r->objht.slots);
  if (r->definitive >= 0)
  {
    r->chainindex = nfields;
    if (r->fields[r->definitive].type == FT_INT)
      r->objht.instr = 1;  /* Hash direct on integer */
  }
  ADD_HASH(recbyname, r);
  return r;
}

static void read_record_desc(void)
{
  /* Read record descriptor.  db is positioned at the opening '['. 
     On return, db points after end of desc (but not after '\n') */
  char * n = read_stringfield(0, 1);	/* Name */
  char * rs = read_stringfield(0, 1);	/* Recognition string */
  int rsl = strlen(rs);
  char * fnp;		/* Field name and parms */
  record * r = NEW(record);
  record * duplicate;
#define MAX_FIELDS 1024
  char * fldnames[MAX_FIELDS];
  int nfields, nf;
  char * fld;
  field * f, * mstrf, * df;
  int got_definitive = 0;
  char * parms;
  char goteol = 0;
  referback * rb;
  char need_definitive = 0;
  unsigned nresrb = 0;
  
  /* Check for valid name */
  if (!valid_name(n))
    DBERROR("Record name not valid (alpha {alphanumeric|underscore}*)");
  
  r->name = n;
  duplicate = find_record(n);	/* Find existing record def of same name */
  if (rsl == 1)
  {
    r->rec1 = rs[0] &= 127;
    if (recognition_2_tab[rs[0]])
      DBERROR("1-char recognition string conflicts with existing 2-char");
  }
  else if (rsl == 2)
  {
    r->rec1 = rs[0] &= 127;
    r->rec2 = rs[1] &= 127;
    if (recognition_1_tab[rs[0]])
      DBERROR("2-char recognition string conflicts with existing 1-char");
  }
  else
    DBERROR("Recognition string not 1 or 2 chars");
  
  /* Search for existing fields have same name as this record name.  If these
     field's record indicates no data read, turn the field into an OBJPTR
     type.  If data has been read, flag error. */
  if (!duplicate)
  for (f = all_fields; f; f = f->next)
    if (!strcmp(f->refname, n))
    {
      if (f->fldno > 31)
        DBERROR("Creating object referral from field after first 32");
      f->type = FT_OBJPTR;
      f->ref = r;
      /* Found a referring field.  Add to this record's referbacks */
      rb = NEW(referback);
      rb->next = r->rbacks;
      rb->fld = f;
      f->rb = rb;
      need_definitive = 1;
      /* Note that objindex and nrbacks are set below when object size 
         is known */
      /*fprintf(stderr, "Record `%s' field %d refers to new `%s' record\n",
        f->owner->name, f->fldno, n);*/
    }
  
  
  nfields = 0;
  while (nfields < MAX_FIELDS)
  {
    fld = fldnames[nfields++] = read_stringfield(0, 1);
    if (fld[0] == ';')
    {
      nfields--;
      break;
    }
  }
  if (nfields == MAX_FIELDS)
    DBERROR("Too many fields.  Probably forgot semicolon");
  if (nfields < 1 && !duplicate)
    DBERROR("No fields defined");
  r->fields = (field *)dbcalloc(sizeof(field) * nfields);
  r->nfields = nfields;
  for (nf = 0, f = r->fields; nf < nfields; nf++, f++)
  {
    fld = fldnames[nf];
    f->owner = r;
    f->fldno = nf;
    if (!strcmp(fld, r->name) ||
        !strncmp(fld, r->name, strlen(r->name)) && fld[strlen(r->name)] == '(')
    {
      while (*fld && fld[1] && fld[1] != '(') fld++;
      fld[0] = '$';
    }
    if (fld[0] == '$')
    {
      if (got_definitive)
        DBERROR("More than one definitive field");
      got_definitive = nf+1;
      f->definitive = 1;
      f->name = f->refname = r->name;
      if (fld[1] == '(')
      {
        fld[1] = 0;
        parms = fld + 2;
      }
      else if (fld[1])
        DBERROR("Bad definitive field specification");
      else
        parms = 0;
    }
    else
    {
      f->name = f->refname = fld;
      if ((parms = strchr(fld, '(')))
        *parms++ = 0;
      if (!valid_name(fld))
        DBERROR("Field name not valid (alpha {alphanumeric|underscore}*)");
    }
    f->type = FT_STRING;
    
    mstrf = find_master_field(fld);
    if (mstrf)
      f->master = mstrf;
    else
    {
      ADD_HASH(&field_by_name, f);
      f->master = f;
    }
    
    if (parms) while (*parms)
    {
      char * p;
      while (isspace(*parms)) parms++;
      if (!*parms || *parms == ')')
          break;
      p = parms;
      while (isalnum(*parms) || *parms == '*' || 
             *parms == '$' || *parms == '=' || *parms == '_') parms++;
      if (*parms == ',' || isspace(*parms)) *parms++ = 0;
      else if (*parms == ')') *parms = 0;
      if (!strcmp(p, "int") || !strcmp(p, "resv"))
      {
        if (f->type != FT_STRING || f->definitive)
          DBERROR("int/resv field cannot be definitive, referring or"
                  " both int and resv");
        f->type = !strcmp(p, "int") ? FT_INT : FT_RESV;
      }
      else if (!strcmp(p, "eol"))
      {
        goteol = 1;
        f->eol = 1;
      }
      else if (!strcmp(p, "nocopy"))
        f->nocopy = 1;
      else if (isdigit(*p) && f->definitive)
      {
        /* Specification of number of reserved referbacks */
        sscanf(p, "%u", &nresrb);
        if (nresrb > 1024)
          DBERROR("Over 1024 reserved referback list slots");
      }
      else if (!strcmp(p, "*"))
        f->implied = 1;
      else if (!strcmp(p, "$"))
        f->objimpl = 1;
      else if (!strncmp(p, "t=", 2))
        f->refname = p+2;
      else
        DBERROR("Unrecognised field parameter");
      if (f->implied && f->objimpl ||
          (f->implied || f->objimpl) && 
          (f->eol || f->type == FT_INT || f->type == FT_RESV))
        DBERROR("* and $ are mutually exclusive with other field parameters");
    }
    
    if (!f->definitive)
    {
      /* See if this field is an existing record datatype_name */
      record * r2 = (record *)lookup_hash(&record_by_name, f->refname);
      if (r2)
      {
        if (r2->definitive < 0)
          DBERROR("Reference to record type with no definitive field");
        f->type = FT_OBJPTR;
        if (nf > 31)
          DBERROR("Object reference field not in first 32");
        f->ref = r2;
        /* If no records of the refered record type have been read, add
           this field as a referback for that record type.  If fields have
           been read, use reserved referback if any.  Else warning. */
        if (!r2->nobj)
        {
          rb = NEW(referback);
          rb->next = r2->rbacks;
          r2->rbacks = rb;
          rb->fld = f;
          f->rb = rb;
          rb->objindex = r2->nfields + r2->nrbacks++;
          r2->objsize += sizeof(void *);
          r2->chainindex++;
          r2->resrback++;
      /*fprintf(stderr, "Record `%s' field %d refers to old `%s' record\n",
        n, f->fldno, r2->name);*/
        }
        else if (r2->resrback && r2->resrback < r2->chainindex)
        {
          /* Target record has reserved referback slot.  Use one. */
          rb = NEW(referback);
          rb->next = r2->rbacks;
          r2->rbacks = rb;
          rb->fld = f;
          f->rb = rb;
          rb->objindex = r2->resrback++;
        }
        else
          DBWARNING("Creating referral to record type which has created "
                    "objects.\n"
                    "No reserved.  Referback lists will not be constructed."); 
      }
    }
  }
  r->definitive = got_definitive - 1;
  if (need_definitive && !got_definitive)
    DBERROR("Previous record definitions refer to this, but no "
            "definitive field specified.");
  
  if (duplicate && nfields)
  {
    /* Duplicate record datatype_name.  Check existing to make sure
       identical. */
    if (r->rec1 != duplicate->rec1 ||
        r->rec2 != duplicate->rec2 ||
        r->nfields != duplicate->nfields)
      DBERROR("Record definition not identical to previous definition");
    for (nf = 0; nf < r->nfields; nf++)
    {
      f = r->fields + nf;
      df = duplicate->fields + nf;
      if (f->master != df->master ||
          f->type != df->type ||
          f->definitive != df->definitive ||
          f->eol != df->eol ||
          f->implied != df->implied ||
          f->objimpl != df->objimpl)
        DBERROR("Record definition not identical to previous definition");
    }
    /* Is identical, discard this definition */
    free(r->fields);
    free(r);
  }
  else
  {
    /* Add to recognition table if not duplicate */
    if (rsl == 1)
    {
      if (recognition_1_tab[rs[0]])
        DBERROR("Recognition string already used");
      recognition_1_tab[rs[0]] = duplicate ? duplicate : r;
    }
    else if (rsl == 2)
    {
      if (recognition_2_tab[rs[0]] &&
          recognition_2_tab[rs[0]][rs[1]])
        DBERROR("Recognition string already used");
      if (!recognition_2_tab[rs[0]])
        recognition_2_tab[rs[0]] = (record **)dbcalloc(sizeof(record *)*128);
      recognition_2_tab[rs[0]][rs[1]] = duplicate ? duplicate : r;
    }
    
    if (duplicate)
      return;
    
    /* Compute object parameters */
    r->objsize = sizeof(object) + sizeof(char *) * 
                 (r->nfields - 1 + !!got_definitive);
    r->objht.slots = OBJ_SLOTS;
    r->objht.offs = sizeof(object) + sizeof(char *) * (r->definitive - 1);
    r->objht.htes = (ht_entry *)dbcalloc(sizeof(ht_entry) * r->objht.slots);
    
    /* Link all fields into all_fields list */
    for (nf = 0; nf < r->nfields; nf++)
    {
      r->fields[nf].next = all_fields;
      all_fields = r->fields + nf;
    }
    
    /* Resolve referbacks to object index */
    for (rb = r->rbacks; rb; rb = rb->next)
    {
      r->objsize += sizeof(char *);
      rb->objindex = r->nfields + r->nrbacks++;
    }
    
    r->resrback = r->nfields + r->nrbacks;
    r->nrbacks += nresrb;
    r->objsize += sizeof(char *) * nresrb;
    
    if (got_definitive)
      r->chainindex = r->nfields + r->nrbacks;
    
    ADD_HASH(&record_by_name, r);
    r->next = all_records;
    all_records = r;
    
    /*fprintf(stderr, "Added record `%s' - %d fields, %d rbacks, key %d\n",
      n, r->nfields, r->nrbacks, r->definitive);*/
      
  }
}

object * create_object(record * r, int nfields, .../* Field values */)
{
  object * o;
  object * mstr = 0;
  va_list ap;
  int fn;
  field * f;
  char * s;
  
  va_start(ap, nfields);
  if (setjmp(jb))
  {
    return 0;
  }
  o = (object *)dbcalloc(r->objsize);
  r->nobj++;
  for (fn = 0, f = r->fields; fn < nfields; fn++, f++)
  {
    s = va_arg(ap, char *);
    o->data[fn] = s;
    if (f->definitive)
    {
      mstr = (object *)lookup_hash(&r->objht, 
                         f->type == FT_INT ? (char *)&s : s);
      if (mstr)
      {
        /* Duplicate record (as defined by definitive field).  Link this
           record to the chain of dup-key records.  The chain pointer is the
           last data element of this object. */
        o->data[r->chainindex] = mstr->data[r->chainindex];
        mstr->data[r->chainindex] = (char *)o;
      }
      else
      {
        ADD_HASH(&r->objht, o);
        r->nuniq++;
      }
      r->last = mstr ? mstr : o;
    }
  }
  if (!mstr)
  {
    o->next = r->objs;
    r->objs = o;
  }
  return o;
}

static void read_record(record * r)
{
  /* Read record.  db is positioned at start of record data.
     On return, db points after end of data (but not after '\n') */
  object * o = (object *)dbcalloc(r->objsize);
  object * mstr = 0;	/* 'Master' object i.e. first object with same key */
  unsigned nf;
  field * f;
  char * fld;
  int spec;
  char resolved;
  
  r->nobj++;
  o->resolved = 0;
  for (nf = 0; nf < r->nfields; nf++)
  {
    f = r->fields + nf;
    if (f->type == FT_RESV)
      continue;
    if (f->implied)
    {
      finish_push();
      fld = "*";
      spec = 1;
    }
    else if (f->objimpl)
    {
      finish_push();
      fld = "$";
      spec = 1;
    }
    else if (f->eol)
      fld = read_to_eol(&spec, f->nocopy);
    else
      fld = read_stringfield(&spec, f->nocopy);
    resolved = 0;
    if (spec)
    {
      if (fld[0] == '*')
        fld = f->master->last, resolved = f->master->last_resolved;
      else if (!f->ref)
      {
        if (f->objimpl)
        {
          if (f->definitive)
            DBERROR("implied $ used in definitive field");
          else
            DBERROR("implied $ used in non-referring field");
        }
        else
        {
          if (f->definitive)
            DBERROR("$ used in definitive field");
          else
            DBERROR("$ used in non-referring field");
        }
      }
      else
        fld = (char *)f->ref->last, resolved = 1;
      if (!fld)
      {
        if (f->implied)
        {
          if (!f->ref)
            DBERROR(
              "implied (*) field, but no explicit field value read previously");
          fld = (char *)f->ref->last, resolved = 1;
        }
        if (!fld)
          DBERROR("* or $: no previous value read");
      }
      discard_last();
    }
    f->master->last = fld;
    f->master->last_resolved = resolved;
    o->data[nf] = fld;
    if (f->definitive)
    {
      mstr = (object *)lookup_hash(&r->objht, fld);
      if (mstr)
      {
        /* Duplicate record (as defined by definitive field).  Link this
           record to the chain of dup-key records.  The chain pointer is the
           last data element of this object. */
        o->data[r->chainindex] = mstr->data[r->chainindex];
        mstr->data[r->chainindex] = (char *)o;
      }
      else
      {
        ADD_HASH(&r->objht, o);
        r->nuniq++;
      }
      r->last = mstr ? mstr : o;
    }
    if (f->type == FT_INT)
    {
      int v;
      sscanf(o->data[nf], "%d", &v);
      o->data[nf] = (char *)v;
    }
    else if (resolved)
      o->resolved |= 1 << nf;
    else if (f->type == FT_OBJPTR)
    {
      /* See if can resolve referred object, otherwise leave as string.
         Doing this now allows us to discard the string if, as in the
         usual case, the referred object has been seen. */
      object * o2 = (object *)lookup_hash(&f->ref->objht, fld);
      if (o2)
      {
        o->data[nf] = (char *)o2;
        o->resolved |= 1 << nf;
        f->master->last = (char *)o2;
        f->master->last_resolved = 1;
        if (!spec)
          discard_last();
      }
    }
    
  }
  
  if (!mstr)
  {
    o->next = r->objs;
    r->objs = o;
  }
  
  for (nf = 0; nf < r->nfields; nf++)
  {
    f = r->fields + nf;
    if (f->rb && o->resolved & 1 << nf)
    {
      /* 'twas resolved to object address.  If referback info available,
         add this object to that object's list of referbacks to this
         object type (whew!) */
      object * o2 = (object *)o->data[nf];
      object_list ** ol = (object_list **)(o2->data + f->rb->objindex);
      object_list * l = NEW(object_list);
      l->next = *ol;
      *ol = l;
      l->obj = mstr ? mstr : o;
    }
  }
  
#ifdef DEBUG
  fprintf(stderr, "Record type `%s':\n", r->name);
  for (nf = 0; nf < r->nfields; nf++)
  {
    f = r->fields + nf;
    fld = o->data[nf];
    if (f->type == FT_INT)
      fprintf(stderr, "  `%s' = %d\n",
        f->name, (int)fld);
    else if (o->resolved & 1 << nf)
      fprintf(stderr, "  `%s' resolved to object `%s'\n",
        f->name, ((object *)fld)->data[f->ref->definitive]);
    else
      fprintf(stderr, "  `%s' = `%s'\n",
        f->name, fld);
  }
#endif
}

int init_db(hash_table ** recbyname)
{
  char_stack = stack_top = char_start = dbmalloc(chunksize = STACK_CHUNK);
  if (!record_by_name.htes)
  {
    record_by_name.slots = RBN_SLOTS;
    record_by_name.offs = offsetof(record, name);
    record_by_name.htes = (ht_entry *)dbcalloc(sizeof(ht_entry)*RBN_SLOTS);
  }
  if (!field_by_name.htes)
  {
    field_by_name.slots = FBN_SLOTS;
    field_by_name.offs = offsetof(field, name);
    field_by_name.htes = (ht_entry *)dbcalloc(sizeof(ht_entry)*FBN_SLOTS);
  }
  if (recbyname)
    *recbyname = &record_by_name;
  return 0;
}


int read_db(char * filename,
            int * numrectypes,
            hash_table ** recbyname,
            record *** recordtab
            )
{
  /* Read filename, returning 0 if no errors, else non-zero.
     If successful, returns hash table which allows access to each object
     type by name, plus an array of pointers to record descriptors. 
     From the record descriptors, the individual objects may be accessed.
  */
  int c, d;
  int sj;
  int ret = 0;
  
  current_file = filename ? filename : "<stdin>";
  
  if (filename)
  {
    db = fopen(filename, "r");
    if (!db)
    {
      fprintf(stderr, "Could not open database file `%s' for reading\n",
        filename);
      return 1;
    }
  }
  else
    db = stdin;
  lineno = 0;
  
  if (!record_by_name.htes)
    init_db(recbyname);
  
  if ((sj = setjmp(jb)))
  {
    fprintf(stderr, "%s:%d: %s\n",
      current_file,
      lineno,
      (char *)sj);
    ret = 1;
    goto cleanup;
  }
  
  for(;;)
  {
    lineno++;
    c = getc(db);
    if (c == EOF)
      break;
    c &= 127;
    if (c == '[')
    {
      ungetc(c, db);
      read_record_desc();
    }
    else if (recognition_1_tab[c])
      read_record(recognition_1_tab[c]);
    else if (recognition_2_tab[c])
    {
      d = getc(db);
      if (d == EOF)
        break;
      d &= 127;
      if (d == '\n')
        continue;
      if (recognition_2_tab[c][d])
        read_record(recognition_2_tab[c][d]);
    }
    while (c != EOF && c != '\n') c = getc(db);
  }

cleanup:
  if (filename)
    fclose(db);

  
  if (numrectypes)
    *numrectypes = record_by_name.nent;
  if (recbyname)
    *recbyname = &record_by_name;
  if (recordtab && record_by_name.nent)
  {
    record * r;
    record ** rp;
    object * o, * mstr, * o2;
    object_list * l, ** ol;
    int nf;
    field * f;
    
    *recordtab = rp = 
      (record **)dbmalloc(sizeof(record *) * record_by_name.nent);
    for (r = all_records; r; r = r->next, rp++)
    {
      *rp = r;
      if (verbose)
      {
        fprintf(stderr, 
          "Record `%s' - fields:%d  rbacks:%d  key field:%d\n"
          "  objects created: %d",
          r->name, r->nfields, r->nrbacks, r->definitive, r->nobj);
        if (r->definitive >= 0)
          fprintf(stderr, "  (%d unique)\n", r->nuniq);
        else
          fputc('\n', stderr);
      }
      
      /* Examine all objects and resolve any resolvable fields that were
         not able to be resolved when they were read (e.g. they referred to
         an object that had not yet been read). */
      for (mstr = r->objs; mstr; mstr = mstr->next)
      {
        for (o = mstr; o;)
        {
          for (nf = 0; nf < r->nfields; nf++)
          {
            f = r->fields + nf;
            if (f->ref && !(o->resolved & 1 << nf))
            {
              /* Found unresolved ref.  See if ref'd object was created */
              o2 = (object *)lookup_hash(&f->ref->objht, o->data[nf]);
              if (o2)
              {
                /* Yes, found it.  Resolve and add referback */
                o->data[nf] = (char *)o2;
                o->resolved |= 1 << nf;
                ol = (object_list **)(o2->data + f->rb->objindex);
                l = NEW(object_list);
                l->next = *ol;
                *ol = l;
                l->obj = mstr;
              }
            }
          }
          if (r->chainindex)
            o = (object *)o->data[r->chainindex];
          else
            break;	/* Not a chainable type */
        }
      }
    }
  }
  return ret;
}
