#ifndef DGDB_H_Incl
#define DGDB_H_Incl
/*****************************************************************************
*   $Id: dgdb.h,v 1.1.1.1 2000/06/22 15:38:24 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgdb.h,v $
* Revision 1.1.1.1  2000/06/22 15:38:24  tekno
* Opened source
*
*
*****************************************************************************/
/* dgdb.h - docgen database structs + hash tables */

#define NAME_LEN 32	/* Max length of datatype and field names */

typedef struct ht_entry_t
{
  struct ht_entry_t * next;	/* For collision resolution */
  char * data;			/* Actual data (needs cast). */
} ht_entry;

typedef struct hash_table_t
{
  unsigned slots;	/* Number of slots (prime number) */
  unsigned nent;	/* Current number of entries used (may be > slots) */
  char instr;		/* Offset below is of 1st char of string (else is
  			   offset of pointer to string) */
  int offs;		/* Offset of string (char * or char[]) in data.  This
  				string is used to compute the hash value */
  ht_entry * htes;	/* Array of hash table entries, dimensioned 'slots'. */
} hash_table;

typedef enum field_type_t
{
  FT_INT,		/* Integer value */
  FT_STRING,		/* String pointer */
  FT_OBJPTR,		/* Pointer to other object */
  FT_RESV		/* Reserved field for later use */
} field_type;

typedef struct field_t
{
  /* Description of record fields */
  char * name;		/* Name of field */
  char * refname;	/* Usually same as 'name', but may be overridden
  			   by 't=recname' parm to indicate record type
  			   referred to. */
  struct record_t * owner;	/* Owner of this field */
  struct field_t * next;	/* Next field in complete list */
  unsigned fldno;		/* Field number (0, 1, ...) */
  field_type type;	/* Type (FT_*) */
  struct field_t * master;	/* Master field (may be self) */
  char definitive;	/* This is definitive object identifier field */
  char eol;		/* This is last field, and data extends to EOL */
  char implied;		/* This field is implied as if '*' */
  char objimpl;		/* This field is implied as if '$' */
  struct record_t * ref;	/* This field refers to other object type.
  				   'ref' is non-null for FT_OBJPTR type. */
  struct referback_t * rb;	/* If above not null, may point to rb
  					list element.  Null if no
  					referback. */
  char nocopy;		/* True if no field copy i.e. '*' as field value
                           stands for itself. */
  char last_resolved;	/* True if following is resolved to object ptr */
  char * last;		/* If this is a master field, point to last value
  				of this field that was read */
} field;

typedef struct object_t
{
  struct object_t * next;
  unsigned long resolved;	/* Bits set 1 for resolved object pointers */
  char * data[1];	/* 1st of arbitrary number of fields.  Each field is
  				either a pointer, or an int (in which case
  				a cast is required in this array).  If
  				resolved, points to another object.
  				Elements beyond this point to lists of
  				objects which refer to this one (one for
  				each object type).
  				If this object type has a definitive field,
  				Element index [nfields+nrbacks] points to the
  				next (duplicate definitive field) instance
  				of this object type.  */
} object;

typedef struct object_list_t
{
  struct object_list_t * next;
  object * obj;
} object_list;

typedef struct referback_t
{
  struct referback_t * next;	/* Next in linked list */
  field * fld;			/* Field which refers to this */
  unsigned objindex;		/* Index in object.data of referback list
  				   for this field's reference */
} referback;

typedef struct record_t
{
  /* Description of each record datatype */
  char * name;		/* Name of datatype */
  struct record_t * next;  /* Next in list of all record types */
  char rec1, rec2;	/* Recognition characters. rec2 may be null.
  			   rec1 may be null if this is array record. */
  unsigned nfields;	/* Number of fields */
  int definitive;	/* Index of definitive field in following array,
  				-1 if no definitive field. */
  field * fields;	/* Array of fields, dimensioned nfields. */
  unsigned objsize;	/* Size (bytes) of objects of this type */
  object * objs;	/* List of objects constructed from database */
  long nobj;		/* Number of objects read so far */
  long nuniq;		/* Number of master (unique definitive) objects */
  hash_table objht;	/* Hash table for accessing above objects by name */
  object * last;	/* Point to last object of this record type read */
  unsigned nrbacks;	/* Number of additional object fields used for
  			   lists of referring objects (referbacks).  Number
  			   of elements in following list.  This number includes
  			   reserved referbacks. */
  referback * rbacks;	/* Referback linked list.  Does not include
  			   reserved referbacks (they are created as needed). */
  unsigned chainindex;	/* Index of object.data used to link object with
  			   same definitive field value.  0 if no linking. */
  unsigned resrback;	/* Index of first reserved referback list.  0 if
  			   no reserved referbacks.  Number of reserved
  			   referbacks can be computed as chainindex-resrback. */
} record;


void add_hash(hash_table * ht, char * data);
void unhash(hash_table * ht, char * data);
void delete_ht(hash_table ** htp, void (*callback)(char *));
hash_table * init_ht(hash_table * ht, unsigned slots, int offs);
char * lookup_hash(hash_table * ht, char * s);
void rehash_table(hash_table * ht, unsigned sl);
int init_db(hash_table ** recbyname);
int read_db(char * filename,
            int * numrectypes,
            hash_table ** recbyname,
            record *** recordtab
            );

record * create_record_type(hash_table * recbyname, char * name,
                            int nfields, ... /* field names and types */);
#define CRT_INT 0x0001  /* Int field (else string) */
#define CRT_DEFINITIVE 0x0002  /* Definitive field (else ordinary data) */
object * create_object(record * r, int nfields, .../* Field values */);

#define ADD_HASH(ht, data) add_hash((ht), (char *)(data))

extern  hash_table * recht;
extern  int nrec;
extern  record ** rpt;

#endif
