/*****************************************************************************
*   $Id: dgfilt.c,v 1.1.1.1 2000/06/22 15:38:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    techno@spirit.net.au
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: dgfilt.c,v $
* Revision 1.1.1.1  2000/06/22 15:38:44  tekno
* Opened source
*
*
*****************************************************************************/
/* dgfilt.c - handle filtering of output.

  Output filters are set up via the outfile statement:
  
    outfile [method_name [ '(' parms ')' ]]... ';'
  
  i.e. a list of zero or more method names, each with optional parameters.
  
  Each of the named methods should be defined as a string method, which
  returns a (possibly transformed) string.
  
  For each line of output which is generated (delimited by newline), the
  first method is invoked for this line.  This method may modify the line;
  delete it or split it into several lines.  To delete the line, the
  method should return a null string.  To split the line, one or more
  newlines should be inserted.  The method should return the line unaltered
  if no change is desired.
  
  The result from the first method is sent to the second method, and so
  on.  The last method's return value is the final output.  If a prior
  method splits a line, subsequent methods are invoked separately for each
  line i.e. each method is given a complete line including newline character.
  If a prior method removes a trailing newline, then subsequent methods
  will not be invoked until a newline has been returned.
  
  Before final output (to stdout), any single leading control character
  (as defined by m_isctrl()) is removed.  The exception to this is if
  there is no filtering defined.
  
  All methods need to implement the following functionality: if passed
  a null string, the filter must flush all pending output that it may have
  been buffering.  If the filter method has no pending data, it should
  return a null string.  This flush operation is called immediately before
  the interpreter terminates i.e. at EOF, and also before a new outfile
  statement is executed.  During a flush, a trailing newline is added to
  returned strings if necessary.  Thus, a flush will always cause at least
  one newline to be output.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <sys/types.h>
#include "dglex.h"
#include "dgvar.h"

#include "dgfilt.h"

#define FILT_BUFSIZE 1024 /* Initial filter buffer size (may be expanded) */

/* Private filtering structures */
typedef struct _filt
{
  /* Individual filter method */
  struct _filt * next;
  int alloc;	/* Size of allocated buffer */
  char * buf;	/* Buffer for input string(s) */
  int used;	/* Used bytes in above buffer */
  tree m;	/* T_METHOD node from outfile statement */
  struct tree_t str;  /* Dummy node for string subject (T_STRING) */
} filt;

typedef struct _fstack
{
  /* Filter stack.  Represents state of all filters specified in most
     recent outfile statement */
  int nf;	/* Current number of filters */
  filt * first;	/* First filter */
  filt * last;
} fstack;


static fstack fs;
static int recurse;  /* Set >0 in filter() */

extern int m_isctrl(char * s);


void filter_int(int i)
{
  char s[20];
  sprintf(s, "%d", i);
  filter(s);
}

static int cat_filt(filt * f, char * s)
{
  int len = strlen(s);
  if (f->used + len >= f->alloc)
  {
      /* Realloc an additional amount of buffer storage */
      f->buf = (char *)realloc(f->buf, f->alloc = f->used + len + FILT_BUFSIZE);
  }
  strcpy(f->buf + f->used, s);
  f->used += len;
  return len;
}

void filter(char * s)
{
  int ei;
  char * es;
  expr_result er;
  filt * f;
  int len;
  char save;
  char * p;
  
  /* Have to prevent inadvertent recursion caused by filter function
     itself trying to output text.  Such text is discarded. */
  if (recurse)
    return;
  
  if (fs.nf < 1)
  {
    fputs(s, outfile);
    return;
  }
  recurse++;
  cat_filt(fs.first, s);
  for (f = fs.first; f; f = f->next)
  {
    /* Check for complete lines.  If any, call filter for each line */
    f->m->u.meth.obj = &f->str;
    s = p = f->buf;
    while (*p && *p != '\n') p++;
    if (!*p)
      break;  /* No complete line; break out */
    while (*p == '\n')
    {
      save = p[1];
      p[1] = 0;  /* Temp null termination */
      f->str.u.strval = s;
      er = invoke_method(&ei, &es, f->m);
      if (er == E_STRING)
      {
        /* Concatenate into next filter's input, or print if last filter */
        if (!f->next)
          fputs(es + (m_isctrl(es) ? 1 : 0), outfile);
        else
          cat_filt(f->next, es);
      }
      p[1] = save;
      p++;
      s = p;
      while (*p && *p != '\n') p++;
    }
    memmove(f->buf, s, p-s);
    f->used = p-s;
    f->buf[f->used] = 0;
  }
  recurse--;
}

void delete_filters(void)
{
  filt * f, * f2;
  
  if (recurse)
    return;
  
  for (f = fs.first; f; f = f2)
  {
    f2 = f->next;
    if (f->buf)
      free(f->buf);
    free(f);
  }
  fs.nf = 0;
  fs.first = fs.last = 0;
}

static filt * create_filt(int bufsize)
{
  filt * f;
  if (recurse)
    return;
  f = (filt *)calloc(sizeof(filt), 1);
  f->buf = (char *)malloc(f->alloc = bufsize);
  f->used = 0;
  f->buf[0] = 0;
  return f;
}

void append_filter(tree m)
{
  if (recurse)
    return;
  if (!fs.nf)
    fs.first = fs.last = create_filt(FILT_BUFSIZE);
  else
  {
    fs.last->next = create_filt(FILT_BUFSIZE);
    fs.last = fs.last->next;
  }
  fs.nf++;
  fs.last->m = m;
  fs.last->str.lineno = m->lineno;
  fs.last->str.filename = m->filename;
  fs.last->str.type = T_STRING;
}

void flush_filters(void)
{
  int ei;
  char * es;
  expr_result er;
  filt * f;
  int len;
  char save;
  char * p;
  char * s;
  
  if (recurse)
    return;
  if (fs.nf < 1)
    return;
  recurse++;

  for (f = fs.first; f; f = f->next)
  {
    s = p = f->buf;
    f->m->u.meth.obj = &f->str;
    if (*s)
    {
      if (f->buf[f->used-1] != '\n')
        cat_filt(f, "\n");
      while (*p && *p != '\n') p++;
      while (*p == '\n')
      {
        save = p[1];
        p[1] = 0;  /* Temp null termination */
        f->str.u.strval = s;
        er = invoke_method(&ei, &es, f->m);
        if (er == E_STRING)
        {
          /* Concatenate into next filter's input, or print if last filter */
          if (!f->next)
            fputs(es + (m_isctrl(es) ? 1 : 0), outfile);
          else
            cat_filt(f->next, es);
        }
        p[1] = save;
        p++;
        s = p;
        while (*p && *p != '\n') p++;
      }
    }
    
    /* Final 'flush' invocation with zero-length string */
    f->used = 0;
    f->buf[0] = 0;
    f->str.u.strval = f->buf;
    er = invoke_method(&ei, &es, f->m);
    if (er == E_STRING)
    {
      if (!f->next)
        fputs(es + (m_isctrl(es) ? 1 : 0), outfile);
      else
        cat_filt(f->next, es);
    }
  }
  recurse--;
}
